from .turbo_mode import TurboMode
from .turbo_optimizer import TurboOptimizer
from .turbo_utils import Telemetry

Turbo = TurboOptimizer

__all__ = [
    "TurboMode",
    "TurboOptimizer",
    "Turbo",
    "Telemetry",
]
